// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2018 Nadav Amit <namit@vmware.com>
 */

#include "gcc-common.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info kernexec_plugin_info = {
	.version	= "201607271510vanilla",
	.help		= "method=call\tmaniuplation method\n"
};

static bool include_emitted;

#define N_CLOBBERED_FUNC_REGS			(4)

struct reg_pair {
	machine_mode	mode;
	unsigned int	regno;
};

static const struct reg_pair clobbered_func_regs[N_CLOBBERED_FUNC_REGS] = {
	{DImode, R11_REG},
	{DImode, R10_REG},
	{CCmode, FLAGS_REG},
	{CCFPmode, FPSR_REG}
};

struct output_pair {
	const char *constraint;
	unsigned int regno;
};

#define N_OUTPUT_FUNC_REGS			(7)

/* VREG indicates the call register, which is N_OUTPUT_FUNC_REGS + 1 */
#define VREG					"8"

static const struct output_pair output_regs[N_OUTPUT_FUNC_REGS] = {
	/* Order must not be changed, since inputs regard outputs */
	{"=r", SP_REG},
	{"=D", DI_REG},
	{"=S", SI_REG},
	{"=c", CX_REG},
	{"=d", DX_REG},
	{"+r", R8_REG},
	{"+r", R9_REG}
};

#define KERNEL_RESTARTABLE_PREFIX		"0x40"

/*
 * %V8, since 8 = N_OUTPUT_FUNC_REGS + 1
 *
 * There are a few suboptimization in this code, that can be addressed in the
 * future. They simplify the code, though.
 *
 * 1. We always encode a longer version of CMP, even 'cmp eax, imm' is possible.
 * 2. We always encode the "restartable" prefix, even on non-preemptive or
 *    voluntary-preemption kernels.
 */
const char *call_block =
	"# INDIRECT BRANCH -------------------				\n"
	"	i = 0							\n"
	"	.irp reg_it, rax,rcx,rdx,rbx,rsp,rbp,rsi,rdi,r8,r9,r10,r11,r12,r13,r14,r15\n"
	"	.ifc \"%V" VREG "\", \"\\reg_it\"			\n"
	"		reg_num=i					\n"
	"	.endif							\n"
	"	i = i + 1						\n"
	"	.endr							\n"
	"1:								\n"
	".section .optpolines,\"a\"					\n"
	"	.quad		1b					\n"
	"	.byte		reg_num					\n"
	".previous							\n"
	"								\n"
	"	.byte 0x48 | ((reg_num & 8) >> 3)			\n"
	"	.byte 0x81, 0xf8 | (reg_num & 7)			\n"
	"	.long 0							\n"
	"								\n"
	"	# jmp 4f, patched to jnz in runtime			\n"
	"	.byte " KERNEL_RESTARTABLE_PREFIX ", 0xeb, 4f - 2f	\n"
	"								\n"
	"	# call retpoline, tell objtool about it			\n"
	"2:								\n"
	"	.pushsection .discard.ignore				\n"
	"	.long 2b - .						\n"
	"	.popsection						\n"
	"	.byte " KERNEL_RESTARTABLE_PREFIX ", 0xe8		\n"
	"	.long __x86_indirect_thunk_%V " VREG " - 3f		\n"
	"3:								\n"
	"	# jmp 5f,  tell objtool about it			\n"
	"	.pushsection .discard.ignore				\n"
	"	.long 3b - .						\n"
	"	.popsection						\n"
	"	.byte 0xeb, 5f - 4f					\n"
	"4:								\n"
	"	# retpoline						\n"
	"	.byte " KERNEL_RESTARTABLE_PREFIX ", 0xe8		\n"
	"	.long __x86_indirect_thunk_%V" VREG " - 5f 		\n"
	"5:								\n"
	" # ----------------------------------				\n";

static unsigned int x86_call_markup_execute(void)
{
	rtx_insn *insn;
	rtx annotate;
	const char *buf;
	const char * name;

	insn = get_first_nonnote_insn();
	if (!insn)
		return 0;

	/* Do not patch init (and other) section calls */
	if (current_function_decl) {
	       const char *sec_name = DECL_SECTION_NAME(current_function_decl);

	       if (sec_name)
		       return 0;
	}

	buf = call_block;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		unsigned int i, j, n_inputs;
		bool has_output;
		rtvec arg_vec, constraint_vec, label_vec;
		rtx operands, call, call_op, annotate;
		rtx asm_op, new_body, p, clob;
		rtx output_reg;
		rtx body;

		if (!CALL_P(insn))
			continue;

		body = PATTERN(insn);
		switch (GET_CODE(body)) {
		case CALL:
			/* A call with no return value */
			has_output = false;
			call = body;
			break;
		case SET:
			/* A call with a return value */
			has_output = true;
			call = SET_SRC(body);
			break;
		default:
			return -1;
		}

		if (GET_CODE(call) != CALL)
			continue;

		call_op = XEXP(XEXP(call, 0), 0);

		switch (GET_CODE(call_op)) {
		case SYMBOL_REF:
			/* direct call */
			continue;
		case REG:
			break;
		default:
			return -1;	/* ERROR */
		}

		/* Count the inputs */
		for (n_inputs = 0, p = CALL_INSN_FUNCTION_USAGE (insn); p; p = XEXP (p, 1)) {
			if (GET_CODE (XEXP (p, 0)) != USE)
				return -1;
			n_inputs++;
		}

		label_vec = rtvec_alloc(0);
		arg_vec = rtvec_alloc(2 + n_inputs);
		constraint_vec = rtvec_alloc(2 + n_inputs);

		i = 0;

		/* AX input */
		RTVEC_ELT(arg_vec, i) = call_op;
		RTVEC_ELT(constraint_vec, i) =
			gen_rtx_ASM_INPUT_loc(GET_MODE(call_op), "r",
					      RTL_LOCATION(call_op));
		i++;

		/* SP input */
		RTVEC_ELT(arg_vec, i) = gen_rtx_REG(DImode, SP_REG);
		RTVEC_ELT(constraint_vec, i) =
			gen_rtx_ASM_INPUT_loc(DImode, "1",
					      RTL_LOCATION(call_op));
		i++;

		for (p = CALL_INSN_FUNCTION_USAGE(insn); p; p = XEXP (p, 1)) {
			const char *constraint;
			rtx input;

			if (GET_CODE (XEXP (p, 0)) != USE)
				continue;

			input = XEXP(XEXP(p, 0), 0);

			if (MEM_P(input)) {
				constraint = "m";
			} else if (REG_P(input)) {
				switch (REGNO(input)) {
				case DI_REG:
					constraint = "D";
					break;
				case SI_REG:
					constraint = "S";
					break;
				case DX_REG:
					constraint = "d";
					break;
				case CX_REG:
					constraint = "c";
					break;
				case R8_REG:
					constraint = "r";
					break;
				case R9_REG:
					constraint = "r";
					break;
				default:
					return -1;
				}
			} else {
				return -1;
			}
			RTVEC_ELT(arg_vec, i) = input;
			rtx input_rtx = gen_rtx_ASM_INPUT_loc(GET_MODE(input),
							      ggc_strdup(constraint),
							      RTL_LOCATION(input));

			RTVEC_ELT(constraint_vec, i) = input_rtx;
			i++;
		}

		new_body = gen_rtx_PARALLEL(VOIDmode,
				rtvec_alloc(1 + 1 + N_OUTPUT_FUNC_REGS +
					    N_CLOBBERED_FUNC_REGS));

		/*
		 * The function output. If none still mark as if AX is
		 * written to ensure it is clobbered.
		 */
		i = 0;
		output_reg = has_output ? SET_DEST(body) :
					  gen_rtx_REG(DImode, AX_REG);
		asm_op = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup(buf), "=a", i,
					      arg_vec, constraint_vec,
					      label_vec, RTL_LOCATION(insn));
		XVECEXP(new_body, 0, i++) = gen_rtx_SET(output_reg, asm_op);

		/*
		 * SP is used as output. Since there is always an output, we do
		 * not use MEM_VOLATILE_P
		 */
		for (j = 0; j < N_OUTPUT_FUNC_REGS; j++) {
			const struct output_pair *output = &output_regs[j];
			rtx reg_rtx;

			asm_op = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup(buf),
						      output->constraint, i,
						      arg_vec, constraint_vec,
						      label_vec, RTL_LOCATION(insn));

			reg_rtx = gen_rtx_REG(DImode, output->regno);
			XVECEXP(new_body, 0, i++) = gen_rtx_SET(reg_rtx, asm_op);
		}

		/* Add the clobbers */
		for (j = 0; j < N_CLOBBERED_FUNC_REGS; j++) {
			const struct reg_pair *regs = &clobbered_func_regs[j];

			clob = gen_rtx_REG(regs->mode, regs->regno);
			clob = gen_rtx_CLOBBER(VOIDmode, clob);
			XVECEXP(new_body, 0, i++) = clob;
		}

		/* Memory clobber */
		clob = gen_rtx_SCRATCH(VOIDmode);
		clob = gen_rtx_MEM(BLKmode, clob);
		clob = gen_rtx_CLOBBER(VOIDmode, clob);
		XVECEXP(new_body, 0, i++) = clob;

		if (n_inputs >= 5)
			emit_insn_before(gen_rtx_USE(VOIDmode,
					  gen_rtx_REG(DImode, R8_REG)), insn);
		if (n_inputs >= 6)
			emit_insn_before(gen_rtx_USE(VOIDmode,
					  gen_rtx_REG(DImode, R9_REG)), insn);

		emit_insn_before(new_body, insn);

		delete_insn(insn);
	}
	return 0;
}

#define PASS_NAME x86_call_markup
#define NO_GATE

#include "gcc-generate-rtl-pass.h"

__visible int plugin_init(struct plugin_name_args *plugin_info,
			  struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument *argv = plugin_info->argv;

	if (!plugin_default_version_check(version, &gcc_version)) {
		error(G_("incompatible gcc/plugin versions"));
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &kernexec_plugin_info);

	PASS_INFO(x86_call_markup, "expand", 1, PASS_POS_INSERT_AFTER);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL,
			  &x86_call_markup_pass_info);

	return 0;
}
